@@ -23,10 +23,10 @@
 #include <unistd.h>
 #include <string.h>
 #include <signal.h>
-typedef void (*sighandler_t)(int);
 #include <sysexits.h>
 #include <sys/time.h>
 #include <pthread.h>
+#include <stdbool.h>
 
 #ifdef __APPLE__
 #include <PCSC/wintypes.h>
@@ -39,22 +39,9 @@ typedef void (*sighandler_t)(int);
 #include "config.h"
 #endif
 
-#ifndef TRUE
-#define TRUE 1
-#define FALSE 0
-#endif
-
-typedef enum
-{
-	True = TRUE, False = FALSE
-} Boolean;
-
 #define TIMEOUT 3600*1000	/* 1 hour timeout */
 
 
-/* command used to parse (on screen) the ATR */
-#define ATR_PARSER "ATR_analysis"
-
 #ifndef SCARD_E_NO_READERS_AVAILABLE
 #define SCARD_E_NO_READERS_AVAILABLE 0x8010002E
 #endif
@@ -96,7 +83,7 @@ do { \
 
 static void usage(const char *pname)
 {
-	printf("%s usage:\n\n\t%s [ -h | -V | -n | -r | -c | -s | -t secs ]\n\n", pname, pname);
+	printf("%s usage:\n\n\t%s [ -h | -V | -n | -r | -c | -s | -t secs | -d | -p]\n\n", pname, pname);
 	printf("\t\t  -h : this help\n");
 	printf("\t\t  -V : print version number\n");
 #ifndef WIN32
@@ -108,6 +95,8 @@ static void usage(const char *pname)
 	printf("\t\t  -q : quiet mode\n");
 	printf("\t\t  -v : verbose mode (default)\n");
 	printf("\t\t  -t secs : quit after secs seconds\n");
+	printf("\t\t  -d : debug mode\n");
+	printf("\t\t  -p : force use of PnP mechanism\n");
 	printf("\n");
 }
 
@@ -121,17 +110,19 @@ const char *cub3 = "";
 const char *cpl = "";
 
 time_t start_time;
-_Atomic Boolean Interrupted = False;
+_Atomic bool Interrupted = false;
 
 typedef struct
 {
 	const char *pname;
-	Boolean analyse_atr;
-	Boolean stress_card;
-	Boolean print_version;
-	Boolean verbose;
-	Boolean only_list_readers;
-	Boolean only_list_cards;
+	bool analyse_atr;
+	bool stress_card;
+	bool print_version;
+	bool verbose;
+	bool only_list_readers;
+	bool only_list_cards;
+	bool debug;
+	bool pnp;
 	long maxtime; // in seconds
 } options_t;
 
@@ -139,7 +130,7 @@ static options_t Options;
 
 SCARDCONTEXT hContext;
 
-static Boolean is_member(const char *  item, const char * list[])
+static bool is_member(const char *  item, const char * list[])
 {
 	int i = 0;
 	while (list[i] && 0 != strcmp(item, list[i]))
@@ -184,24 +175,24 @@ static void initialize_terminal(void)
 }
 /* There should be no \033 beyond this line! */
 
-static Boolean should_exit(void)
+static bool should_exit(void)
 {
 	if (Options.maxtime)
 	{
 		time_t t = time(NULL);
 		if (t - start_time > Options.maxtime)
-			return True;
+			return true;
 	}
 
 #ifdef WIN32
 	if (GetKeyState(VK_SHIFT) & 0x80)
-		return True;
+		return true;
 #endif
 
 	if (Interrupted)
-		return True;
+		return true;
 
-	return False;
+	return false;
 }
 
 typedef enum
@@ -237,7 +228,7 @@ static void *spin_update(void *p)
 	(void)p;
 
 again:
-	/* wait untill spinning starts */
+	/* wait until spinning starts */
 	do
 	{
 		if (should_exit())
@@ -279,7 +270,7 @@ again:
 static void user_interrupt_signal_handler(int signal)
 {
 	(void)signal;
-	Interrupted = True;
+	Interrupted = true;
 }
 
 static void initialize_signal_handlers(void)
@@ -291,19 +282,23 @@ static void initialize_signal_handlers(void)
 static void initialize_options(options_t *options, const char *pname)
 {
 	options->pname = pname;
-#if defined(WIN32) || defined(__APPLE__)
-	options->analyse_atr = False;
+#if defined(ATR_PARSER)
+	options->analyse_atr = true;
 #else
-	options->analyse_atr = True;
+	options->analyse_atr = false;
+#define ATR_PARSER ""
 #endif
-	options->stress_card = False;
+	options->stress_card = false;
+	options->print_version = false;
+	options->verbose = true;
+	options->only_list_readers = false;
+	options->only_list_cards = false;
+	options->debug = false;
+	options->pnp = false;
 	options->maxtime = 0;
-	options->verbose = True;
-	options->only_list_readers = False;
-	options->only_list_cards = False;
 }
 
-#define OPTIONS_BASE "Vhvrcst:q"
+#define OPTIONS_BASE "Vhvrcst:qdp"
 #ifdef WIN32
 #define OPTIONS OPTIONS_BASE
 #else
@@ -327,7 +322,7 @@ static int parse_options(int argc, char *argv[], options_t *options)
 		switch (opt)
 		{
 			case 'n':
-				options->analyse_atr = False;
+				options->analyse_atr = false;
 				break;
 
 			case 'V':
@@ -336,22 +331,22 @@ static int parse_options(int argc, char *argv[], options_t *options)
 				break;
 
 			case 'v':
-				options->verbose = True;
+				options->verbose = true;
 				break;
 
 			case 'r':
-				options->only_list_readers = True;
-				options->verbose = False;
+				options->only_list_readers = true;
+				options->verbose = false;
 				break;
 
 			case 'c':
-				options->only_list_cards = True;
-				options->verbose = False;
-				options->analyse_atr = False;
+				options->only_list_cards = true;
+				options->verbose = false;
+				options->analyse_atr = false;
 				break;
 
 			case 's':
-				options->stress_card = True;
+				options->stress_card = true;
 				break;
 
 			case 't':
@@ -359,13 +354,21 @@ static int parse_options(int argc, char *argv[], options_t *options)
 				break;
 
 			case 'q':
-				options->verbose = False;
+				options->verbose = false;
+				break;
+
+			case 'p':
+				options->pnp = true;
 				break;
 
 			case 'h':
 				usage(pname);
 				exit(EX_OK);
 
+			case 'd':
+				options->debug = true;
+				break;
+
 			default:
 				usage(pname);
 				exit(EX_USAGE);
@@ -480,6 +483,45 @@ static void print_readers(const char **readers, int nbReaders)
 	}
 }
 
+static void displayChangedStatus(SCARD_READERSTATE rgReaderStates[], int count)
+{
+	char state_bits[][sizeof "UNAVAILABLE"] = {
+		"IGNORE",		/* 0x0001 */
+		"CHANGED",		/* 0x0002 */
+		"UNKNOWN",		/* 0x0004 */
+		"UNAVAILABLE",	/* 0x0008 */
+		"EMPTY",		/* 0x0010 */
+		"PRESENT",		/* 0x0020 */
+		"ATRMATCH",		/* 0x0040 */
+		"EXCLUSIVE",	/* 0x0080 */
+		"INUSE",		/* 0x0100 */
+		"MUTE",			/* 0x0200 */
+		"UNPOWERED"		/* 0x0400 */
+	};
+	printf("\n");
+	for (int i=0; i<count; i++)
+	{
+		SCARD_READERSTATE r = rgReaderStates[i];
+		printf("%d: %s, %d, 0x%04X -> 0x%04X",
+			i,
+			r.szReader,
+			(int)(r.dwEventState >> 16),
+			(int)(r.dwCurrentState & 0xFFFF),
+			(int)(r.dwEventState & 0xFFFF));
+		for (int b=0; b<11; b++)
+		{
+			int v = 1 << b;
+			if ((r.dwEventState & v) && (r.dwCurrentState & v))
+				printf(" =%s", state_bits[b]);
+			if ((r.dwEventState & v) && !(r.dwCurrentState & v))
+				printf(" %s+%s%s", blue, state_bits[b], color_end);
+			if (!(r.dwEventState & v) && (r.dwCurrentState & v))
+				printf(" %s-%s%s", red, state_bits[b], color_end);
+		}
+		printf("\n");
+	}
+}
+
 int main(int argc, char *argv[])
 {
 	int current_reader;
@@ -497,7 +539,6 @@ int main(int argc, char *argv[])
 	int nbReaders, i;
 	char atr[MAX_ATR_SIZE*3+1];	/* ATR in ASCII */
 	char atr_command[sizeof(atr)+sizeof(ATR_PARSER)+2+1];
-	int pnp = TRUE;
 	pthread_t spin_pthread;
 
 	start_time = time(NULL);
@@ -527,16 +568,16 @@ int main(int argc, char *argv[])
 	rgReaderStates[0].dwCurrentState = SCARD_STATE_UNAWARE;
 
 	rv = SCardGetStatusChange(hContext, 0, rgReaderStates, 1);
-	if (rgReaderStates[0].dwEventState & SCARD_STATE_UNKNOWN)
+	if (! Options.pnp && rgReaderStates[0].dwEventState & SCARD_STATE_UNKNOWN)
 	{
 		if (Options.verbose)
 		{
 			printf("%sPlug'n play reader name not supported. Using polling every %d ms.%s\n", magenta, TIMEOUT, color_end);
 		}
-		pnp = FALSE;
 	}
 	else
 	{
+		Options.pnp = true;
 		if (Options.verbose)
 		{
 			printf("%sUsing reader plug'n play mechanism%s\n", magenta, color_end);
@@ -618,7 +659,7 @@ get_readers:
 			fflush(stdout);
 		}
 
-		if (pnp)
+		if (Options.pnp)
 		{
 			rgReaderStates[0].szReader = "\\\\?PnP?\\Notification";
 			rgReaderStates[0].dwCurrentState = SCARD_STATE_UNAWARE;
@@ -632,6 +673,9 @@ get_readers:
 			spin_stop();
 
 			test_rv("SCardGetStatusChange", rv, hContext);
+
+			if (Options.debug)
+				displayChangedStatus(rgReaderStates, 1);
 		}
 		else
 		{
@@ -707,16 +751,16 @@ get_readers:
 	}
 
 	/* If Plug and Play is supported by the PC/SC layer */
-	if (pnp)
+	if (Options.pnp)
 	{
 		rgReaderStates_t[nbReaders].szReader = "\\\\?PnP?\\Notification";
-		rgReaderStates_t[nbReaders].dwCurrentState = SCARD_STATE_UNAWARE;
+		rgReaderStates_t[nbReaders].dwCurrentState = SCARD_STATE_UNKNOWN;
 		nbReaders++;
 	}
 
 #ifdef WIN32
 	int oldNbReaders;
-	int oldNbReaders_init = FALSE;
+	int oldNbReaders_init = false;
 #endif
 	spin_start();
 
@@ -727,11 +771,13 @@ get_readers:
 
 	spin_stop();
 
+	if (Options.debug)
+		displayChangedStatus(rgReaderStates_t, nbReaders);
 	while ((rv == SCARD_S_SUCCESS) || (rv == SCARD_E_TIMEOUT))
 	{
 		time_t t;
 
-		if (pnp)
+		if (Options.pnp)
 		{
 			/* check if the number of readers has changed */
 #ifdef WIN32
@@ -739,7 +785,7 @@ get_readers:
 			if (! oldNbReaders_init)
 			{
 				oldNbReaders = newNbReaders;
-				oldNbReaders_init = TRUE;
+				oldNbReaders_init = true;
 			}
 			if (newNbReaders != oldNbReaders)
 #else
@@ -777,15 +823,12 @@ get_readers:
 				continue;
 #endif
 
-			if (rgReaderStates_t[current_reader].dwEventState &
-				SCARD_STATE_CHANGED)
-			{
-				/* If something has changed the new state is now the current
-				 * state */
-				rgReaderStates_t[current_reader].dwCurrentState =
-					rgReaderStates_t[current_reader].dwEventState;
-			}
-			else
+			/* The new current state is now the old event state */
+			rgReaderStates_t[current_reader].dwCurrentState =
+				rgReaderStates_t[current_reader].dwEventState;
+
+			if (! (rgReaderStates_t[current_reader].dwEventState &
+				SCARD_STATE_CHANGED))
 				/* If nothing changed then skip to the next reader */
 				continue;
 
@@ -800,8 +843,8 @@ get_readers:
 				color_end);
 
 			/* Event number */
-			printf("  Event number: %s%ld%s\n", magenta,
-				rgReaderStates_t[current_reader].dwEventState >> 16,
+			printf("  Event number: %s%d%s\n", magenta,
+				(int)(rgReaderStates_t[current_reader].dwEventState >> 16),
 				color_end);
 
 			/* Dump the full current state */
@@ -881,6 +924,8 @@ get_readers:
 					sprintf(atr_command, ATR_PARSER " '%s'", atr);
 					if (system(atr_command))
 						perror(atr_command);
+
+					printf("\n");
 				}
 			}
 
@@ -907,6 +952,9 @@ get_readers:
 			nbReaders);
 
 		spin_stop();
+
+		if (Options.debug)
+			displayChangedStatus(rgReaderStates_t, nbReaders);
 	} /* while */
 
 	/* A reader disappeared */
